/*
 * PM2: Parallel Multithreaded Machine
 * Copyright (C) 2001 "the PM2 team" (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */


#include "marcel.h"
#include <errno.h>


#ifdef MA__IFACE_LPT
int lpt_condattr_init(lpt_condattr_t * attr)
{
	MARCEL_LOG_IN();
	memset(attr, '\0', sizeof(*attr));
	MA_BUG_ON(sizeof(lpt_condattr_t) > __SIZEOF_LPT_CONDATTR_T);
	MARCEL_LOG_RETURN(0);
}
DEF_LIBPTHREAD(int, condattr_init, (pthread_condattr_t * attr), (attr))
DEF___LIBPTHREAD(int, condattr_init, (pthread_condattr_t * attr), (attr))

int lpt_condattr_destroy(lpt_condattr_t * attr TBX_UNUSED)
{
	MARCEL_LOG_IN();
	/* Nothing to be done.  */
	MARCEL_LOG_RETURN(0);
}
DEF_LIBPTHREAD(int, condattr_destroy, (pthread_condattr_t * attr), (attr))
DEF___LIBPTHREAD(int, condattr_destroy, (pthread_condattr_t * attr), (attr))

int lpt_condattr_getpshared(const lpt_condattr_t * __restrict attr, int *__restrict pshared)
{
	MARCEL_LOG_IN();
	*pshared = ((const struct lpt_condattr *) attr)->value & 1;
	MARCEL_LOG_RETURN(0);
}
DEF_LIBPTHREAD(int, condattr_getpshared, (pthread_condattr_t * __restrict attr, int *__restrict pshared), (attr, pshared))

int lpt_condattr_setpshared(lpt_condattr_t * attr, int pshared)
{
	MARCEL_LOG_IN();
	if (pshared != LPT_PROCESS_PRIVATE  && __builtin_expect(pshared != LPT_PROCESS_SHARED, 0)) {
		MARCEL_LOG("lpt_condattr_setpshared : valeur pshared(%d)  invalide\n", pshared);
		MARCEL_LOG_RETURN(EINVAL);
	}

	/* For now it is not possible to share a mutex variable.  */
	if (pshared != LPT_PROCESS_PRIVATE) {
		MA_NOT_SUPPORTED("shared condition");
		MARCEL_LOG_RETURN(ENOTSUP);
	}

	int *valuep = &((struct lpt_condattr *) attr)->value;
	*valuep = (*valuep & ~1) | (pshared != LPT_PROCESS_PRIVATE);

	MARCEL_LOG_RETURN(0);
}
DEF_LIBPTHREAD(int, condattr_setpshared, (pthread_condattr_t * attr, int pshared), (attr, pshared))


int lpt_cond_init(lpt_cond_t * __restrict cond, const lpt_condattr_t * __restrict attr TBX_UNUSED)
{
	const struct _lpt_fastlock initlock = MA_LPT_FASTLOCK_UNLOCKED;

	MARCEL_LOG_IN();

	cond->__data.__lock = initlock;
	cond->__data.__waiting = NULL;

	MARCEL_LOG_RETURN(0);
}
versioned_symbol(libpthread, lpt_cond_init, pthread_cond_init, GLIBC_2_3_2);

int lpt_cond_destroy(lpt_cond_t * cond)
{
	MARCEL_LOG_IN();

	if (cond->__data.__waiting != NULL)
		MARCEL_LOG_RETURN(EBUSY);
	__lpt_destroy_lock(&cond->__data.__lock);

	MARCEL_LOG_RETURN(0);
}
versioned_symbol(libpthread, lpt_cond_destroy, pthread_cond_destroy, GLIBC_2_3_2);

int lpt_cond_signal(lpt_cond_t * cond)
{
	MARCEL_LOG_IN();
	__lpt_unlock(&cond->__data.__lock);
	MARCEL_LOG_RETURN(0);
}
versioned_symbol(libpthread, lpt_cond_signal, pthread_cond_signal, GLIBC_2_3_2);

int lpt_cond_broadcast(lpt_cond_t * cond)
{
	MARCEL_LOG_IN();

	lpt_fastlock_acquire(&cond->__data.__lock);
	while (__lpt_unlock_spinlocked(&cond->__data.__lock));
	lpt_fastlock_release(&cond->__data.__lock);

	MARCEL_LOG_RETURN(0);
}
versioned_symbol(libpthread, lpt_cond_broadcast, pthread_cond_broadcast, GLIBC_2_3_2);

int lpt_cond_wait(lpt_cond_t * __restrict cond, lpt_mutex_t * __restrict mutex)
{
	MARCEL_LOG_IN();

	/** release the mutex and wait on cond */
	lpt_fastlock_acquire(&mutex->__data.__lock);
	lpt_fastlock_acquire(&cond->__data.__lock);
	mutex->__data.__owner = 0;
	__lpt_unlock_spinlocked(&mutex->__data.__lock);
	lpt_fastlock_release(&mutex->__data.__lock);
	__lpt_lock_wait(&cond->__data.__lock, ma_self(), MA_CHECK_CANCEL);
	lpt_fastlock_release(&cond->__data.__lock);

	/** reacquire the mutex and check is I was canceled */
	lpt_mutex_lock(mutex);
#ifdef MARCEL_DEVIATION_ENABLED
	if (SELF_GETMEM(canceled) == MARCEL_IS_CANCELED)
                marcel_exit(PMARCEL_CANCELED);
#endif
	MARCEL_LOG_RETURN(0);
}
versioned_symbol(libpthread, lpt_cond_wait, pthread_cond_wait, GLIBC_2_3_2);

int lpt_cond_timedwait(lpt_cond_t * __restrict cond, lpt_mutex_t * __restrict mutex,
		       const struct timespec *__restrict abstime)
{
	MARCEL_LOG_IN();

	struct timeval now, tv;
	unsigned long timeout;
	int ret;

	tv.tv_sec = abstime->tv_sec;
	tv.tv_usec = abstime->tv_nsec / 1000;
	gettimeofday(&now, NULL);
	if (timercmp(&tv, &now, <=))
		MARCEL_LOG_RETURN(ETIMEDOUT);
	timeout = MA_TIMEVAL_TO_USEC(&tv) - MA_TIMEVAL_TO_USEC(&now);

	/** release the mutex and wait on cond */
	lpt_fastlock_acquire(&mutex->__data.__lock);
	lpt_fastlock_acquire(&cond->__data.__lock);
	mutex->__data.__owner = 0;
	__lpt_unlock_spinlocked(&mutex->__data.__lock);
	lpt_fastlock_release(&mutex->__data.__lock);
	ret = __lpt_lock_timed_wait(&cond->__data.__lock, ma_self(), timeout, MA_CHECK_CANCEL);

	/** reacquire the mutex and check is I was canceled */
	lpt_mutex_lock(mutex);
#ifdef MARCEL_DEVIATION_ENABLED
        if (SELF_GETMEM(canceled) == MARCEL_IS_CANCELED)
                marcel_exit(PMARCEL_CANCELED);
#endif
	MARCEL_LOG_RETURN(ret);
}
versioned_symbol(libpthread, lpt_cond_timedwait, pthread_cond_timedwait, GLIBC_2_3_2);
#endif


#if defined(MA__LIBPTHREAD) && MA_GLIBC_VERSION_MINIMUM < 20302
DEF_STRONG_ALIAS(lpt_cond_init, __old_pthread_cond_init)
compat_symbol(libpthread, __old_pthread_cond_init, pthread_cond_init, GLIBC_2_0);
DEF_STRONG_ALIAS(lpt_cond_destroy, __old_pthread_cond_destroy)
compat_symbol(libpthread, __old_pthread_cond_destroy, pthread_cond_destroy, GLIBC_2_0);
DEF_STRONG_ALIAS(lpt_cond_signal, __old_pthread_cond_signal)
compat_symbol(libpthread, __old_pthread_cond_signal, pthread_cond_signal, GLIBC_2_0);
DEF_STRONG_ALIAS(lpt_cond_broadcast, __old_pthread_cond_broadcast)
compat_symbol(libpthread, __old_pthread_cond_broadcast, pthread_cond_broadcast, GLIBC_2_0);
DEF_STRONG_ALIAS(lpt_cond_wait, __old_pthread_cond_wait)
compat_symbol(libpthread, __old_pthread_cond_wait, pthread_cond_wait, GLIBC_2_0);
DEF_STRONG_ALIAS(lpt_cond_timedwait, __old_pthread_cond_timedwait)
compat_symbol(libpthread, __old_pthread_cond_timedwait, pthread_cond_timedwait, GLIBC_2_0);
#endif
